import logging

from netbox import settings
from netbox_zabbix.jsonrpc import JSONRPC

logger = logging.getLogger('netbox.plugins.netbox_zabbix')


class Zabbix:
    jsonrpc = None

    url = None
    username = None
    password = None

    def __init__(self):

        if not settings.PLUGINS_CONFIG.get('netbox_zabbix', {}).get('url', None):
            logger.error(f'Zabbix: No Zabbix URL configured')
            return False
        elif not settings.PLUGINS_CONFIG.get('netbox_zabbix', {}).get('username', None):
            logger.error(f'Zabbix: No Zabbix username configured')
            return False
        elif not settings.PLUGINS_CONFIG.get('netbox_zabbix', {}).get('password', None):
            logger.error(f'Zabbix: No Zabbix password configured')
            return False
        else:
            self.url = settings.PLUGINS_CONFIG.get('netbox_zabbix', {}).get('url', None)
            self.username =  settings.PLUGINS_CONFIG.get('netbox_zabbix', {}).get('username', None)
            self.password = settings.PLUGINS_CONFIG.get('netbox_zabbix', {}).get('password', None)

        self.jsonrpc = JSONRPC(url=self.url, username=self.username, password=self.password)

    def template_get(self, name):
        data = {
            'method': 'template.get',
            'params': {
                "output": [
                    "hostid"
                ],
                "filter": {
                    "host": name
                }
            },
        }

        response = self.jsonrpc.send_api_request(data)
        result = response.json()

        if len(result.get('result', [])) > 1:
            raise Exception('Too many hosts found')
        elif len(result.get('result', [])) == 0:
            return None

        return result.get('result', []).pop()

    @staticmethod
    def build_interface(snmp, ip):
        if snmp:
            interfaces = []
            interface_type = 2
            main = 1
            port = '161'
            if snmp.get('version', None) == '2' or snmp.get('version', None) == '2c':
                interfaces.append({
                    'type': interface_type,
                    'main': main,
                    'useip': 1,
                    'ip': ip,
                    'dns': '',
                    'port': port,
                        'details': {
                        'version': 2,
                        'bulk': 0,
                        'community': '{$SNMP_COMMUNITY}'
                    }
                })

            if len(interfaces) > 0:
                return {'interfaces': interfaces}
            else:
                return {}
        else:
            return {}

    @staticmethod
    def build_macro(name, value):
        if name and value:
            logger.debug(f'\tZabbix: Building Macros')
            return {'macros': [
                        {
                            'macro': '{$' + name + '}',
                            'value': f'{value}'
                        }
                    ]}
        else:
            return {}

    def hostinterface_get(self, hostid):
        data = {
            'method': 'hostinterface.get',
            'params': {
                'hostids': [hostid]
            },
        }

        response = self.jsonrpc.send_api_request(data)
        result = response.json()

        if len(result.get('result', [])) > 1:
            raise Exception('Too many hosts found')
        elif len(result.get('result', [])) == 0:
            return None

        return result.get('result', []).pop()

    def hostinterface_update(self, hostid, interfaceid, ip=None):
        data = {
            'method': 'hostinterface.update',
            'params': {
                'interfaceid': interfaceid
            },
        }
        if ip:
            data['params'].update({'ip': f'{ip}'})
        response = self.jsonrpc.send_api_request(data)
        result = response.json()

    def host_get(self, host=None, hostid=None):
        if hostid:
            data = {
                'method': 'host.get',
                'params': {
                    'hostids': [int(hostid)]
                },
            }
        else:
            data = {
                'method': 'host.get',
                'params': {
                    "filter": {
                        "host": host
                    }
                },
            }

        response = self.jsonrpc.send_api_request(data)
        result = response.json()

        if len(result.get('result', [])) > 1:
            raise Exception('Too many hosts found')
        elif len(result.get('result', [])) == 0:
            return None

        return result.get('result', []).pop()

    def host_create(self, name, ip, templates, groups, type=2, main=1, port=161, snmp={}, status=0):
        host = self.host_get(host=name)

        if not host:
            data = {
                'method': 'host.create',
                'params': {
                    'host': name,
                    'groups': groups,
                    'templates': templates,
                    "inventory_mode": 1,
                    "status": status,
                },
            }
            if ip and snmp:
                data['params'].update(self.build_interface(snmp=snmp, ip=ip))
            if snmp:
                community = snmp.get('community')
                data['params'].update(self.build_macro('SNMP_COMMUNITY', community))

            response = self.jsonrpc.send_api_request(data)
            result = response.json()
        else:
            raise Exception('Cannot Create: Host Exists')

    def host_update(self, hostid, name, ip, templates, groups, type=2, main=1, port=161, snmp={}, status=0):
        host = None
        if hostid:
            host = self.host_get(hostid=hostid)
        else:
            host = self.host_get(host=name)

        if host:
            interface = self.hostinterface_get(hostid=hostid)
            data = {
                'method': 'host.update',
                'params': {
                    'hostid': f'{hostid}',
                    'host': name,
                    'templates': templates,
                    'groups': groups,
                    'status': status,
                },
            }
            if snmp:
                data['params'].update(self.build_macro('SNMP_COMMUNITY', snmp.get('community', None)))
            response = self.jsonrpc.send_api_request(data)
            result = response.json()

            if ip and interface.get('ip', None) != ip:
                # Only update interface if IP doesn't match
                self.hostinterface_update(hostid=hostid, interfaceid=interface.get('interfaceid'), ip=ip)
        else:
            self.host_create(name, ip, templates, groups, type, main, port, snmp)



import json
import logging
import requests

logger = logging.getLogger('netbox.plugins.netbox_zabbix')


class JSONRPC:

    url = None
    username = None
    password = None
    token = None
    id = 1

    ssl_verify = False
    ca_file_path = None
    proxies = None

    def __init__(self, url, username, password, ssl_verify=False, ca_file_path=None, proxies=None):
        self.url = url
        self.username = username
        self.password = password
        self.ssl_verify = ssl_verify
        self.ca_file_path = ca_file_path
        self.proxies = proxies

        if not self.logged_in():
            self.login()

    def send_api_request(self, data):
        if not data.get('id', None):
            data['id'] = self.id
        if not data.get('token', None):
            data['auth'] = self.token
        if not data.get('jsonrpc', None):
            data['jsonrpc'] = '2.0'

        headers = {
            'Content-Type': 'application/json',
        }
        body = json.dumps(data, cls=json.JSONEncoder)
        # Prepare the HTTP request
        params = {
            'method': 'POST',
            'url': self.url,
            'headers': headers,
            'data': body.encode('utf8'),
        }

        try:
            prepared_request = requests.Request(**params).prepare()
        except requests.exceptions.RequestException as e:
            logger.error(f"JSONRPC: Error forming HTTP request: {e}")
            raise e

        with requests.Session() as session:
            session.verify = self.ssl_verify
            if self.ca_file_path:
                session.verify = self.ca_file_path

            response = session.send(prepared_request, proxies=self.proxies)
            self.id += 1

        if 200 <= response.status_code <= 299:
            if response.json().get('error', {}).get('message', None):
                message = response.json().get('error', {}).get('message', None)
                error_data = response.json().get('error', {}).get('data', None)
                logger.debug(f'JSONRPC Result Payload: {data}')
                raise requests.exceptions.RequestException(
                    f"Error {message} returned with data '{error_data}'.  Response: {response.json()}"
                )
            else:
                # logger.info(f"Request succeeded; response status {response.status_code}")
                return response
        else:
            logger.warning(f"JSONRPC: Request failed; response status {response.status_code}: {response.content}")
            logger.debug(f'JSONRPC Result Payload: {data}')
            raise requests.exceptions.RequestException(
                f"Status {response.status_code} returned with content '{response.content}'"
            )

    def logged_in(self):
        if not self.token:
            return False
        return True

    def login(self):
        data = {
            'jsonrpc': '2.0',
            'method': 'user.login',
            'params': {
                'user': self.username,
                'password': self.password
            },
            'id': self.id,
            'auth': self.token
        }
        response = self.send_api_request(data)

        result = response.json()
        self.token = result.get('result', None)
        return result.get('result', None)

    def logout(self):

        data = {
            'jsonrpc': '2.0',
            'method': 'user.logout',
            'params': [],
            'id': self.id,
            'auth': self.token
        }
        response = self.send_api_request(data)

        result = response.json()
        self.auth = None
        return result.get('result', None)



import logging
import pprint

from django.db.models.signals import post_save, m2m_changed
from django.dispatch import receiver
from django_rq import get_queue

from netbox import settings
from extras.models import Tag, TaggedItem
from dcim.models import Device
from netbox_zabbix.zabbix import Zabbix

logger = logging.getLogger('netbox.plugins.netbox_zabbix')


def can_do_update(instance):
    if instance.get_config_context().get('zabbix', {}).get('groups', None) and \
            instance.device_type.custom_field_data.get('zabbix_group') == '':
        logger.debug(f'Zabbix({instance}): Appropriate groups not set')
        return False

    if not instance.primary_ip:
        logger.debug(f'Zabbix({instance}): Missing primary IP')
        return False

    if not instance.name:
        logger.debug(f'Zabbix({instance}): Missing name')
        return False

    tag_names = settings.PLUGINS_CONFIG.get('netbox_zabbix', {}).get('tags', None)
    logger.debug(f'{tag_names}')
    if tag_names is not None:
        if not isinstance(tag_names, list):
            tag_names = [tag_names]
        tag_ids = Tag.objects.filter(name__in=tag_names).values_list('id', flat=True)
        if instance.tags.filter(pk__in=[tag_ids]).count() == 0:
            logger.debug(f'Zabbix({instance.name}): Device does not match Zabbix automation tags')
            return False

    return True

def snmp_details(device):
    config_context = device.get_config_context()
    if config_context.get('zabbix', {}).get('snmp', None):
        snmp = None
        return config_context.get('zabbix', {}).get('snmp', {})
    else:
        return {}

def update_hostid(zabbix, device, name=None):
    host = zabbix.host_get(host=name)
    if host and host.get('hostid', None):
        device.skip_signal = True
        device.custom_field_data['zabbix_hostid'] = int(host.get('hostid'))
        device.save()
    else:
        logger.error(f'Zabbix({device.name}): Host ID not set and unable to locate host in Zabbix')
        return

def update_zabbix(pk, hostid=None):
    try:
        instance = Device.objects.get(pk=pk)
    except Device.DoesNotExist:
        logger.error(f'Zabbix({pk}): Device Instance not found')
        return

    status = 0
    if instance.status in ['offline', 'planned', 'inventory', 'staged']:
        status = 1

    if hasattr(instance, '_prechange_snapshot') and instance.name != instance._prechange_snapshot.get('name'):
        old_name = instance._prechange_snapshot.get('name')
    else:
        old_name = instance.name
    try:
        zabbix = Zabbix()
        snmp = snmp_details(device=instance)
        hostid = instance.custom_field_data.get('zabbix_hostid', None)

        if not hostid:
            update_hostid(zabbix, instance, old_name)
        elif hostid and not zabbix.host_get(hostid=hostid):
            update_hostid(zabbix, instance, old_name)
            instance.refresh_from_db()
            hostid = instance.custom_field_data.get('zabbix_hostid', None)

        template = zabbix.template_get(instance.device_type.get_full_name)
        if template is None:
            template = zabbix.template_get(f'{instance.device_type.manufacturer.name} {instance.device_type.part_number}')

        group = instance.get_config_context().get('zabbix', {}).get('groups', [])
        group.extend(instance.get_config_context().get('zabbix', {}).get('tenants', []))
        groups = []
        for gid in group:
            groups.append({'groupid': f"{gid}"})
        if len(groups) == 0:
            groups.append({'groupid': f"{settings.PLUGINS_CONFIG.get('netbox_zabbix', {}).get('group', None)}"})
        if template:
            logger.info(f'Zabbix({instance.name}): Starting update')
            zabbix.host_update(
                hostid=instance.custom_field_data.get('zabbix_hostid', None),
                name=instance.name,
                ip=f'{instance.primary_ip.address.ip}',
                templates=[{'templateid': f"{template.get('templateid')}"}],
                groups=groups,
                snmp=snmp,
                status=status,
            )

            if not hostid:
                update_hostid(zabbix, instance, old_name)
            elif hostid and not zabbix.host_get(hostid=hostid):
                update_hostid(zabbix, instance, old_name)

            logger.info(f'Zabbix({instance.name}): Complete for {instance.name}')
        else:
            logger.error(f'Zabbix({instance.name}): No template set ("{instance.device_type.get_full_name}") ')
    except Exception as e:
        import traceback
        logger.error(f'Zabbix({instance.name}): Exception: {e}')
        logger.error(traceback.format_exc())


@receiver(post_save, sender=Device)
def update_device(instance, **kwargs):
    if hasattr(instance, 'skip_signal') and instance.skip_signal:
        return
    if can_do_update(instance):
        logger.debug('NetBox Zabbix: Hit Signal')
        queue = get_queue('high')
        job = queue.enqueue(
            'netbox_zabbix.signals.update_zabbix',
            description=f'zabbix_update-{instance.name}',
            pk=instance.pk,
        )
    else:
        logger.error(f'No update available for {instance}')

@receiver(m2m_changed, sender=TaggedItem)
def m2m_device(instance, **kwargs):
    if kwargs.get('action', None) not in ['post_add'] or not kwargs.get('pk_set') or not isinstance(instance, Device) \
            or hasattr(instance, 'skip_signal') and instance.skip_signal:
        return
    if can_do_update(instance):
        logger.debug('NetBox Zabbix: Hit M2M')
        queue = get_queue('high')
        job = queue.enqueue(
            'netbox_zabbix.signals.update_zabbix',
            description=f'zabbix_update-{instance.name}',
            pk=instance.pk,
        )
    else:
        logger.error(f'No update available for {instance}')



from setuptools import find_packages, setup

setup(
    name='netbox_zabbix',
    version='1.0.0',
    description='NetBox Zabbix Sync',
    long_description='Plugin to Sync to Zabbix',
    url='https://github.com/dansheps/netbox-zabbix/',
    download_url='https://github.com/dansheps/netbox-zabbix/',
    author='Daniel Sheppard',
    author_email='dans@dansheps.com',
    maintainer='Daniel Sheppard',
    maintainer_email='dans@dansheps.com',
    install_requires=[],
    packages=find_packages(),
    include_package_data=True,
    license='GNU GPL v3',
    zip_safe=False,
    keywords=['netbox', 'netbox-plugin'],
    classifiers=[
        'Framework :: Django',
        'Programming Language :: Python :: 3',
    ]
)


from extras.plugins import PluginConfig
from importlib.metadata import metadata

metadata = metadata('netbox_zabbix')

class ZabbixPlugin(PluginConfig):
    name = metadata.get('Name').replace('-', '_')
    verbose_name = metadata.get('Summary')
    description = metadata.get('Description')
    version = metadata.get('Version')
    author = metadata.get('Author')
    author_email = metadata.get('Author-email')
    base_url = 'zabbix'
    min_version = '3.2.0'
    max_version = '3.5.99'
    required_settings = [
        'username',
        'password',
    ]
    default_settings = {
        'tags': ['automation: monitoring'],
        'snmp': {
            'version': '2c',  # 1, 2c, 3
            'community': None,
            'context': None,
            'username': None,
            'level': None,  # None, noAuthNoPriv, authNoPriv, authPriv
            'auth_protocol': None,  # None, MD5, SHA1, SHA224, SHA256, SHA384, SHA512
            'auth_passphrase': None,
            'priv_protocol': None,  # None, DES, AES128, AES192, AES256, AES192C, AES256C
            'priv_passphrase': None,
        }
    }
    queues = []

    def ready(self):
        super().ready()
        import netbox_zabbix.signals


config = ZabbixPlugin
